from django_temporalio.registry import schedules
from temporalio.client import (
    Schedule,
    ScheduleActionStartWorkflow,
    ScheduleIntervalSpec,
    ScheduleSpec,
)

from django_kafka.conf import settings
from django_kafka.relations_resolver.temporalio.workflows import ResolveRelations

schedules.register(
    "resolve-awaited-relations",
    Schedule(
        action=ScheduleActionStartWorkflow(
            ResolveRelations.run,
            id="resolve-awaited-relations",
            task_queue=settings.TEMPORAL_TASK_QUEUE,
        ),
        spec=ScheduleSpec(
            intervals=[
                ScheduleIntervalSpec(every=settings.RELATION_RESOLVER_DAEMON_INTERVAL),
            ],
        ),
    ),
)



from abc import ABC, abstractmethod
from collections.abc import AsyncIterator
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from confluent_kafka import cimpl

    from django_kafka.relations_resolver.relation import Relation


class MessageProcessor(ABC):
    @abstractmethod
    async def aadd_message(self, msg: "cimpl.Message", relation: "Relation"): ...

    @abstractmethod
    async def adelete(self, relation: "Relation") -> bool: ...

    @abstractmethod
    async def aexists(self, relation: "Relation") -> bool: ...

    @abstractmethod
    async def aprocess_messages(self, relation: "Relation"): ...

    @abstractmethod
    async def ato_resolve(self) -> AsyncIterator["Relation"]: ...

    @abstractmethod
    async def amark_resolving(self): ...



from unittest import mock

from django.db.models import Model
from django.test import SimpleTestCase, TestCase

from django_kafka.tests.utils import message_mock
from django_kafka.topic.reproducer import ReproduceTopic, TopicReproducer


class ReproduceTopicTestCase(SimpleTestCase):
    def setUp(self):
        class _Topic(ReproduceTopic):
            reproducer = mock.Mock()
            name = "dummy"

        self.topic = _Topic()

    def test_consume_skips_when_kafka_skip_true_and_not_deletion(self):
        self.topic.deserialize = mock.Mock(
            side_effect=[
                {"id": 1},
                {"kafka_skip": True},
            ],
        )

        self.topic.consume(message_mock())

        self.topic.reproducer.reproduce.assert_not_called()

    def test_consume_calls_reproducer_on_deletion_when_value_is_none(self):
        key = {"id": 123}
        value = None
        self.topic.deserialize = mock.Mock(side_effect=[key, value])

        self.topic.consume(message_mock())

        self.topic.reproducer.reproduce.assert_called_once_with(key, value, True)

    def test_consume_calls_reproducer_on_deletion_flag_true(self):
        key = {"id": 5}
        value = {"__deleted": "true"}
        self.topic.deserialize = mock.Mock(side_effect=[key, value])

        self.topic.consume(message_mock())

        self.topic.reproducer.reproduce.assert_called_once_with(key, value, True)

    def test_consume_calls_reproducer_on_normal_event(self):
        key = {"id": 77}
        value = {"foo": "bar"}
        self.topic.deserialize = mock.Mock(side_effect=[key, value])

        self.topic.consume(message_mock())

        self.topic.reproducer.reproduce.assert_called_once_with(key, value, False)


class TopicReproducerGetTopicTestCase(SimpleTestCase):
    def setUp(self):
        class TestModel(Model):
            def __str__(self):
                return ""

        self.model = TestModel

    def test_get_reproduce_topic_with_name_and_namespace(self):
        class TestTopicReproducer(TopicReproducer):
            reproduce_name = "table"
            reproduce_namespace = "ns"

        topic = TestTopicReproducer.get_reproduce_topic()

        self.assertIsInstance(topic, ReproduceTopic)
        self.assertEqual(topic.name, "ns.table")
        self.assertIsInstance(topic.reproducer, TestTopicReproducer)

    def test_get_reproduce_topic_with_model(self):
        class TestTopicReproducer(TopicReproducer):
            reproduce_model = self.model
            reproduce_namespace = "ns"

        topic = TestTopicReproducer.get_reproduce_topic()

        self.assertIsInstance(topic, ReproduceTopic)
        self.assertEqual(topic.name, f"ns.{self.model._meta.db_table}")
        self.assertIsInstance(topic.reproducer, TestTopicReproducer)

    def test_get_reproduce_topic_raises_when_not_configured(self):
        class TestTopicReproducer(TopicReproducer):
            pass

        with self.assertRaises(ValueError):
            TestTopicReproducer.get_reproduce_topic()


class TopicReproducerBehaviourTestCase(TestCase):
    def setUp(self):
        class TestModel(Model):
            def __str__(self):
                return ""

        self.model = TestModel

    def test_reproduce_calls_deletion_when_is_deletion_true(self):
        class TestTopicReproducer(TopicReproducer):
            reproduce_model = self.model

        r = TestTopicReproducer()
        r._reproduce_deletion = mock.Mock()
        r._reproduce_upsert = mock.Mock()
        key = {"id": 10}
        value = None

        r.reproduce(key, value, is_deletion=True)

        r._reproduce_deletion.assert_called_once_with(10, key, value)
        r._reproduce_upsert.assert_not_called()

    def test_reproduce_calls_upsert_when_instance_exists(self):
        class TestTopicReproducer(TopicReproducer):
            reproduce_model = self.model

        self.model.objects.get = mock.Mock()
        r = TestTopicReproducer()
        r._reproduce_upsert = mock.Mock()
        r._reproduce_deletion = mock.Mock()
        key = {"id": 42}
        value = {"some": "data"}

        r.reproduce(key, value, is_deletion=False)

        self.model.objects.get.assert_called_once_with(id=42)
        r._reproduce_upsert.assert_called_once_with(
            self.model.objects.get.return_value,
            key,
            value,
        )
        r._reproduce_deletion.assert_not_called()

    def test_reproduce_suppresses_when_instance_missing(self):
        class TestTopicReproducer(TopicReproducer):
            reproduce_model = self.model

        self.model.objects.get = mock.Mock(side_effect=self.model.DoesNotExist)
        r = TestTopicReproducer()
        r._reproduce_upsert = mock.Mock()
        r._reproduce_deletion = mock.Mock()
        key = {"id": 1}
        value = {"x": 1}

        r.reproduce(key, value, is_deletion=False)

        self.model.objects.get.assert_called_once_with(id=1)
        r._reproduce_upsert.assert_not_called()
        r._reproduce_deletion.assert_not_called()



from unittest.mock import Mock

from faker import Faker

from django_kafka.utils.message import MessageTimestamp


def message_mock(  # noqa: PLR0913
    topic="topic",
    partition=0,
    offset=0,
    error=None,
    headers=None,
    timestamp: list[MessageTimestamp, int] | None = None,
):
    """mocking utility for confluent_kafka.cimpl.Message"""
    return Mock(
        **{
            "topic.return_value": topic,
            "partition.return_value": partition,
            "offset.return_value": offset,
            "headers.return_value": headers,
            "error.return_value": error,
            "key.return_value": Faker().binary(length=10),
            "timestamp.return_value": timestamp
            or (MessageTimestamp.NOT_AVAILABLE, Faker().unix_time() * 1000),
        },
    )


class AsyncIteratorMock(Mock):
    def __init__(self, items):
        super().__init__()
        self._iter = iter(items)
        self.return_value = self

    def __aiter__(self):
        return self

    async def __anext__(self):
        try:
            return next(self._iter)
        except StopIteration as e:
            raise StopAsyncIteration from e



from datetime import timedelta

from django_temporalio.registry import queue_activities
from django_temporalio.utils import heartbeat
from temporalio import activity

from django_kafka import kafka
from django_kafka.conf import settings
from django_kafka.relations_resolver.daemon import get_daemon
from django_kafka.relations_resolver.relation import RelationType, SerializedRelation


@queue_activities.register(settings.TEMPORAL_TASK_QUEUE)
@activity.defn
async def resolve_relation(serialized_relation: SerializedRelation):
    async with heartbeat(timedelta(seconds=5)):
        await kafka.relations_resolver.aresolve_relation(
            RelationType.instance(serialized_relation),
        )


@queue_activities.register(settings.TEMPORAL_TASK_QUEUE)
@activity.defn
async def check_resolved_relations():
    async with heartbeat(timedelta(seconds=5)):
        await get_daemon().aresolve_relations()



from unittest.mock import AsyncMock, Mock, patch

from django.test import TestCase

from django_kafka.exceptions import TopicNotRegisteredError
from django_kafka.models import WaitingMessage, WaitingMessageQuerySet
from django_kafka.relations_resolver.processor.model import ModelMessageProcessor
from django_kafka.tests.utils import AsyncIteratorMock


class ModelMessageProcessorTestCase(TestCase):
    def setUp(self):
        self.msg_processor = ModelMessageProcessor()

    def test__init__(self):
        self.assertEqual(self.msg_processor.model, WaitingMessage)

    @patch("django_kafka.models.WaitingMessage.objects", spec=WaitingMessageQuerySet)
    @patch(
        "django_kafka.relations_resolver.processor.model.sync_to_async",
        return_value=AsyncMock(),
    )
    async def test_aadd_message(self, mock_sync_to_async, mock_qs):
        msg = Mock()
        relation = Mock()

        await self.msg_processor.aadd_message(msg, relation)

        mock_sync_to_async.assert_called_once_with(mock_qs.add_message)
        mock_sync_to_async().assert_called_once_with(msg, relation)

    @patch("django_kafka.models.WaitingMessage.objects", spec=WaitingMessageQuerySet)
    @patch(
        "django_kafka.relations_resolver.processor.model.sync_to_async",
        return_value=AsyncMock(),
    )
    async def test_adelete(self, mock_sync_to_async, mock_qs):
        relation = Mock()

        await self.msg_processor.adelete(relation)

        mock_sync_to_async.assert_called_once_with(mock_qs.for_relation)
        mock_sync_to_async().assert_called_once_with(relation)
        mock_sync_to_async().return_value.adelete.assert_called_once_with()

    @patch("django_kafka.models.WaitingMessage.objects", spec=WaitingMessageQuerySet)
    @patch(
        "django_kafka.relations_resolver.processor.model.sync_to_async",
        return_value=AsyncMock(),
    )
    async def test_aexists(self, mock_sync_to_async, mock_qs):
        relation = Mock()

        await self.msg_processor.aexists(relation)

        mock_sync_to_async.assert_called_once_with(mock_qs.for_relation)
        mock_sync_to_async().assert_called_once_with(relation)
        mock_sync_to_async().return_value.aexists.assert_called_once_with()

    @patch("django_kafka.models.WaitingMessage.objects", spec=WaitingMessageQuerySet)
    @patch(
        "django_kafka.relations_resolver.processor.model.sync_to_async",
        return_value=AsyncMock(),
    )
    async def test_amark_resolving(self, mock_sync_to_async, mock_qs):
        relation = Mock()

        await self.msg_processor.amark_resolving(relation)

        mock_sync_to_async.assert_called_once_with(mock_qs.mark_resolving)
        mock_sync_to_async().assert_called_once_with(relation)

    async def test_ato_resolve(self):
        not_to_resolve = [
            AsyncMock(**{"aexists.return_value": False}),
        ]
        to_resolve = [
            AsyncMock(**{"aexists.return_value": True}),
            AsyncMock(**{"aexists.return_value": True}),
            AsyncMock(**{"aexists.return_value": True}),
        ]
        relations = [
            *not_to_resolve,
            *to_resolve,
        ]
        qs_items = [Mock(**{"relation.return_value": r}) for r in relations]

        with patch(
            "django_kafka.models.WaitingMessage.objects",
            aiter_relations_to_resolve=AsyncIteratorMock(qs_items),
        ):
            result = [r async for r in self.msg_processor.ato_resolve()]

        for relation in not_to_resolve:
            self.assertNotIn(relation, result)

        for relation in to_resolve:
            self.assertIn(relation, result)

    async def test__aget_missing_relation(self):
        existing_relations = [
            AsyncMock(**{"aexists.return_value": True}),
            AsyncMock(**{"aexists.return_value": True}),
            AsyncMock(**{"aexists.return_value": False}),
            AsyncMock(**{"aexists.return_value": True}),
            AsyncMock(**{"aexists.return_value": False}),
        ]
        topic = Mock(
            **{
                "get_relations.return_value": [
                    *existing_relations,
                ],
            },
        )
        msg = Mock()

        result = await self.msg_processor._aget_missing_relation(topic, msg)

        self.assertEqual(result, existing_relations[2])

    @patch("django_kafka.models.WaitingMessage.objects")
    @patch(
        "django_kafka.relations_resolver.processor.model.ModelMessageProcessor._aget_missing_relation",
    )
    @patch("django_kafka.kafka.consumers.topic")
    @patch("django_kafka.relations_resolver.processor.model.sync_to_async")
    async def test_aprocess_messages_topic_not_registered(
        self,
        mock_sync_to_async,
        mock_consumers_topic,
        mock__aget_missing_relation,
        mock_qs,
    ):
        relation = Mock()
        model_msg = AsyncMock()

        mock_sync_to_async.return_value = AsyncMock(
            return_value=AsyncIteratorMock([model_msg]),
        )
        mock_consumers_topic.side_effect = TopicNotRegisteredError

        await self.msg_processor.aprocess_messages(relation)

        mock_sync_to_async.assert_called_once_with(mock_qs.for_relation)
        model_msg.adelete.assert_called_once_with()
        mock__aget_missing_relation.assert_not_called()

    @patch("django_kafka.models.WaitingMessage.objects")
    @patch(
        "django_kafka.relations_resolver.processor.model.ModelMessageProcessor._aget_missing_relation",
    )
    @patch("django_kafka.kafka.relations_resolver.await_for_relation")
    @patch("django_kafka.kafka.consumers.topic")
    @patch("django_kafka.relations_resolver.processor.model.Message")
    @patch("django_kafka.relations_resolver.processor.model.sync_to_async")
    async def test_aprocess_messages_some_relation_is_missing(
        self,
        mock_sync_to_async,
        mock_kafka_message,
        mock_consumers_topic,
        mock_await_for_relation,
        mock__aget_missing_relation,
        mock_qs,
    ):
        relation = Mock()
        model_msg = AsyncMock()
        missing_relation = Mock()

        mock_sync_to_async.return_value = AsyncMock(
            return_value=AsyncIteratorMock([model_msg]),
        )
        mock__aget_missing_relation.return_value = missing_relation

        await self.msg_processor.aprocess_messages(relation)

        mock_sync_to_async.assert_called_once_with(mock_qs.for_relation)
        mock__aget_missing_relation.assert_called_once_with(
            mock_consumers_topic(),
            mock_kafka_message(),
        )
        mock_await_for_relation.assert_called_once_with(
            mock_kafka_message(),
            missing_relation,
        )
        model_msg.adelete.assert_called_once_with()

    @patch(
        "django_kafka.relations_resolver.processor.model.ModelMessageProcessor._aget_missing_relation",
    )
    @patch("django_kafka.kafka.relations_resolver.await_for_relation")
    @patch("django_kafka.kafka.consumers.topic")
    @patch("django_kafka.relations_resolver.processor.model.Message")
    @patch("django_kafka.relations_resolver.processor.model.sync_to_async")
    async def test_aprocess_messages_consume(
        self,
        mock_sync_to_async,
        mock_kafka_message,
        mock_consumers_topic,
        mock_await_for_relation,
        mock__aget_missing_relation,
    ):
        relation = Mock()
        model_msg = AsyncMock()

        sync_to_async_results = [
            # for model.objects.for_relation() call
            AsyncMock(return_value=AsyncIteratorMock([model_msg])),
            # for topic.consume() call
            AsyncMock(),
        ]

        mock_sync_to_async.side_effect = sync_to_async_results
        mock__aget_missing_relation.return_value = None

        await self.msg_processor.aprocess_messages(relation)

        mock_await_for_relation.assert_not_called()
        mock_sync_to_async.assert_any_call(mock_consumers_topic().consume)
        sync_to_async_results[1].assert_any_call(mock_kafka_message())
        model_msg.adelete.assert_called_once_with()



from unittest import mock

from django.db import models

from django_kafka.connect.models import KafkaConnectSkipModel
from django_kafka.exceptions import DjangoKafkaError
from django_kafka.tests.models import AbstractModelTestCase
from django_kafka.topic.model import ModelTopicConsumer


class ModelTopicConsumerTestCase(AbstractModelTestCase):
    abstract_model = models.Model
    model: type[KafkaConnectSkipModel]

    def _get_model_topic_consumer(self):
        class SomeModelTopicConsumer(ModelTopicConsumer):
            name = "name"
            model = self.model

            def get_lookup_kwargs(self, model, key, value) -> dict:
                return {}

            def is_deletion(self, *args, **kwargs):
                return False

        return SomeModelTopicConsumer()

    def test_get_defaults(self):
        topic_consumer = self._get_model_topic_consumer()

        class SomeModel(models.Model):
            name = models.CharField()

            class Meta:
                abstract = True

        defaults = topic_consumer.get_defaults(
            model=SomeModel,
            value={"name": 1, "not_model_field": "value"},
        )

        self.assertEqual(defaults, {"name": 1})

    def test_get_defaults__adds_kafka_skip(self):
        topic_consumer = self._get_model_topic_consumer()

        class KafkaConnectSkip(KafkaConnectSkipModel):
            name = models.CharField()

        defaults = topic_consumer.get_defaults(
            model=KafkaConnectSkip,
            value={"name": 1},
        )

        self.assertEqual(defaults, {"name": 1, "kafka_skip": True})

    def test_transform(self):
        """test custom transform methods are used during transformation"""
        topic_consumer = self._get_model_topic_consumer()
        topic_consumer.transform_name = mock.Mock(
            return_value=("transformed_name", "transformed_value"),
        )

        transformed_value = topic_consumer.transform(self.model, {"name": "value"})

        self.assertEqual(transformed_value, {"transformed_name": "transformed_value"})
        topic_consumer.transform_name.assert_called_once_with(
            self.model,
            "name",
            "value",
        )

    def test_exclude_fields(self):
        """test excluded fields are not included in the transformed result"""
        topic_consumer = self._get_model_topic_consumer()
        topic_consumer.exclude_fields = ["name"]

        transformed_value = topic_consumer.transform(self.model, {"name": "value"})

        self.assertEqual(transformed_value, {})

    def test_sync(self):
        topic_consumer = self._get_model_topic_consumer()
        topic_consumer.get_lookup_kwargs = mock.Mock(return_value={"id": "id"})
        topic_consumer.transform = mock.Mock()
        topic_consumer.get_defaults = mock.Mock(return_value={"name": "name"})
        model = mock.Mock()

        results = topic_consumer.sync(model, {"key": "key"}, {"value": "value"})

        topic_consumer.get_lookup_kwargs.assert_called_once_with(
            model,
            {"key": "key"},
            {"value": "value"},
        )
        topic_consumer.transform.assert_called_once_with(
            model,
            {"value": "value"},
        )
        topic_consumer.get_defaults.assert_called_once_with(
            model,
            topic_consumer.transform.return_value,
        )
        model.objects.update_or_create.assert_called_once_with(
            id="id",
            defaults={"name": "name"},
        )
        self.assertEqual(results, model.objects.update_or_create.return_value)

    def test_sync__deleted(self):
        topic_consumer = self._get_model_topic_consumer()
        topic_consumer.get_lookup_kwargs = mock.Mock(return_value={"id": "id"})
        topic_consumer.get_defaults = mock.Mock()
        topic_consumer.is_deletion = mock.Mock(return_value=True)
        model = mock.Mock()

        results = topic_consumer.sync(model, {"key": "key"}, {"value": "value"})

        topic_consumer.get_lookup_kwargs.assert_called_once_with(
            model,
            {"key": "key"},
            {"value": "value"},
        )
        topic_consumer.get_defaults.assert_not_called()
        model.objects.get.assert_called_once_with(id="id")
        model.objects.get.return_value.delete.assert_called_once()
        model.objects.update_or_create.assert_not_called()
        self.assertIsNone(results)

    def test_get_model(self):
        topic_consumer = self._get_model_topic_consumer()
        topic_consumer.model = mock.Mock()

        self.assertEqual(
            topic_consumer.get_model({}, {}),
            topic_consumer.model,
        )

    def test_get_model__raises_when_model_not_set(self):
        topic_consumer = self._get_model_topic_consumer()
        topic_consumer.model = None

        with self.assertRaises(DjangoKafkaError):
            topic_consumer.get_model({}, {})

    def test_consume(self):
        topic_consumer = self._get_model_topic_consumer()
        topic_consumer.get_model = mock.Mock()
        msg_key = {"key": "key"}
        msg_value = {"value": "value"}
        topic_consumer.deserialize = mock.Mock(side_effect=[msg_key, msg_value])
        topic_consumer.sync = mock.Mock()

        topic_consumer.consume(mock.Mock())

        topic_consumer.get_model.assert_called_once_with(msg_key, msg_value)
        topic_consumer.sync.assert_called_once_with(
            topic_consumer.get_model.return_value,
            msg_key,
            msg_value,
        )

    def test_model_has_field(self):
        class BookModel(models.Model):
            name = models.CharField(max_length=100)
            author = models.ForeignKey(self.model, models.CASCADE)

            def __str__(self):
                return self.name

        topic_consumer = self._get_model_topic_consumer()
        topic_consumer.model = BookModel

        fields = ("id", "name", "author", "author_id")
        for field in fields:
            self.assertTrue(topic_consumer.model_has_field(BookModel, field))

        self.assertFalse(topic_consumer.model_has_field(BookModel, "not_defined_field"))



from datetime import UTC, datetime
from enum import IntEnum

MessageHeaders = list[tuple[str, str]]


class Header:
    @classmethod
    def list(cls, headers: MessageHeaders | None, header: str) -> list[str]:
        """returns all occurrences for the header"""
        if headers is not None:
            return [v for k, v in headers if k == header]
        return []

    @classmethod
    def get(cls, headers: MessageHeaders | None, header: str) -> str | None:
        """returns the first encountered value for the header"""
        if headers is not None:
            return next((v for k, v in headers if k == header), None)
        return None


class MessageTimestamp(IntEnum):
    """
    https://docs.confluent.io/platform/current/clients/confluent-kafka-python/html/index.html#confluent_kafka.Message.timestamp
    Values are taken from the constants from `confluent_kafka.cimpl.py`
    """

    NOT_AVAILABLE = 0
    CREATE_TIME = 1
    LOG_APPEND_TIME = 2

    @classmethod
    def to_datetime(
        cls,
        msg_timestamp: tuple["MessageTimestamp", int],
    ) -> datetime | None:
        timestamp_type, timestamp = msg_timestamp

        if timestamp_type == MessageTimestamp.NOT_AVAILABLE:
            return None

        return datetime.fromtimestamp(timestamp / 1000, UTC)


class Message:
    """
    This only exists because cimpl.Message is not instantiatable.
    https://github.com/confluentinc/confluent-kafka-python/issues/1535
    """

    # ruff: noqa: PLR0913
    def __init__(
        self,
        topic: str,
        key: bytes | str,
        value: bytes | str | None,
        headers: list[tuple[str, bytes]] | None,
        offset: int,
        partition: int,
        timestamp: tuple[MessageTimestamp, int],
    ):
        self._topic = topic
        self._key = key
        self._value = value
        self._headers = headers
        self._offset = offset
        self._partition = partition
        self._timestamp = timestamp

    def topic(self) -> str:
        return self._topic

    def key(self) -> bytes | str:
        if type(self._key) is list:
            return bytes(self._key)
        return self._key

    def value(self) -> bytes | str:
        if type(self._value) is list:
            return bytes(self._value)
        return self._value

    def headers(self) -> list[tuple[str, bytes]] | None:
        return self._headers

    def timestamp(self) -> tuple[MessageTimestamp, int] | None:
        return self._timestamp

    def partition(self) -> int:
        return self._partition

    def offset(self) -> int:
        return self._offset



from unittest.mock import AsyncMock, MagicMock, Mock, patch

from django.test import SimpleTestCase

from django_kafka.relations_resolver.relation import Relation
from django_kafka.relations_resolver.resolver import RelationResolver
from django_kafka.tests.utils import message_mock


class RelationResolverTestCase(SimpleTestCase):
    async def test_await_for_relation(self):
        resolver = RelationResolver()
        msg = Mock()
        relation = MagicMock(spec=Relation)

        await resolver.await_for_relation(msg, relation)

        relation.aadd_message.assert_called_once_with(msg)

    @patch(
        "django_kafka.relations_resolver.resolver.RelationResolver.await_for_relation",
        new_callable=AsyncMock,
    )
    async def test_aresolve_relation_missing_wait(self, mock_await_for_relation):
        resolver = RelationResolver()
        msg = message_mock()

        relation = MagicMock(spec=Relation)
        # relations missing - send to waiting queue
        relation.aexists.return_value = False

        action = await resolver.aresolve([relation], msg)

        mock_await_for_relation.assert_called_once_with(msg, relation)
        relation.aexists.assert_called_once_with()
        self.assertEqual(action, RelationResolver.Action.SKIP)

        # relations exists and already has waiting messages - pause
        mock_await_for_relation.reset_mock()
        relation.reset_mock()
        relation.aexists.return_value = True
        relation.ahas_waiting_messages.return_value = True

        action = await resolver.aresolve([relation], msg)

        mock_await_for_relation.assert_not_called()
        relation.aexists.assert_called_once_with()
        relation.ahas_waiting_messages.assert_called_once_with()
        self.assertEqual(action, RelationResolver.Action.PAUSE)

        # relation exists, no waiting messages - continue
        mock_await_for_relation.reset_mock()
        relation.reset_mock()
        relation.aexists.return_value = True
        relation.ahas_waiting_messages.return_value = False

        action = await resolver.aresolve([relation], msg)

        mock_await_for_relation.assert_not_called()
        relation.aexists.assert_called_once_with()
        relation.ahas_waiting_messages.assert_called_once_with()
        self.assertEqual(action, RelationResolver.Action.CONTINUE)

    async def test_aresolve_relation(self):
        resolver = RelationResolver()
        relation = MagicMock(spec=Relation)

        await resolver.aresolve_relation(relation)

        relation.aresolve.assert_called_once_with()



import os
import time

import factory
from factory.django import DjangoModelFactory

from django_kafka.models import WaitingMessage
from django_kafka.utils.message import MessageTimestamp
from example.models import Order


class WaitingMessageFactory(DjangoModelFactory):
    class Meta:
        model = WaitingMessage

    key = factory.LazyFunction(lambda: os.urandom(16))
    value = factory.LazyFunction(lambda: os.urandom(16))
    timestamp = factory.LazyAttribute(
        lambda o: (MessageTimestamp.CREATE_TIME.value, int(time.time() * 1000)),
    )
    topic = factory.Faker("uuid4")
    partition = 1
    offset = 1000


class OrderFactory(DjangoModelFactory):
    class Meta:
        model = Order



from django.test import SimpleTestCase

from django_kafka.relations_resolver.relation import ModelRelation, RelationType
from example.models import Order


class RelationTypeTestCase(SimpleTestCase):
    def test_type_model_instance(self):
        serialized_relation = {
            "type": "MODEL",
            "kwargs": {
                "model": ModelRelation.get_model_key(Order),
                "id_value": 1,
                "id_field": "id",
            },
        }

        relation = RelationType.instance(serialized_relation)

        self.assertIsInstance(relation, ModelRelation)
        self.assertEqual(relation.model, Order)
        self.assertEqual(relation.id_field, serialized_relation["kwargs"]["id_field"])
        self.assertEqual(relation.id_value, serialized_relation["kwargs"]["id_value"])
        self.assertEqual(relation.model_key, serialized_relation["kwargs"]["model"])



import contextlib
from abc import ABC
from typing import TYPE_CHECKING, Any

from confluent_kafka.serialization import MessageField

from django_kafka.conf import settings
from django_kafka.topic.avro import AvroTopicConsumer

if TYPE_CHECKING:
    from django.db.models import Model


class ReproduceTopic(AvroTopicConsumer, ABC):
    """
    Class to consume messages from a source topic and distribute them to a specific
    topic reproducer class which augments the events with extra data as needed.

    The typical use case is to consume messages from a debezium source connector and
    then add related field data to those messages.
    """

    reproducer: "TopicReproducer"

    @classmethod
    def _is_deletion(cls, value) -> bool:
        return value is None or value.get("__deleted") == "true"

    @classmethod
    def _skip_reproduce(cls, value):
        # necessary to not filter out deletion events for rows marked as kafka_skip=true
        return not cls._is_deletion(value) and value.get("kafka_skip")

    def consume(self, msg):
        key = self.deserialize(msg.key(), MessageField.KEY, msg.headers())
        value = self.deserialize(msg.value(), MessageField.VALUE, msg.headers())

        if not self._skip_reproduce(value):
            self.reproducer.reproduce(key, value, self._is_deletion(value))


class TopicReproducer:
    """
    Simplifies reproducing messages from Django model instance changes or a custom
    topic. The following attributes/methods may be overridden:

    1. `reproduce_model` (optional) - the model class for which messages will be
        reproduced from events to a debezium source connector topic. If not set,
        then `reproduce_name` must be set and `reproduce` overridden.
    2. `reproduce_name` (optional) - the topic name to reproduce messages from, for when
        there is no `reproduce_model` or a custom topic name is required.
    3. `reproduce_namespace` (optional) - if the debezium source connector prepends
        topic names with a namespace, specify this here.
    4. `reproduce` (optional) - defines default reproduce behaviour which calls
        `_reproduce_upsert` and `_reproduce_deletion` depending on instance upsert or
        deletion respectively. These latter two methods must be implemented if this
        method is not overridden.
    """

    reproduce_namespace: str | None = settings.REPRODUCER_CONFIG["namespace"]
    reproduce_model: type["Model"] | None = None
    reproduce_name: str | None = None

    @classmethod
    def get_reproduce_topic(cls):
        """
        creates a simple default reproducer topic (consumer) for this producer

        for more complex setup, subclass `ReproduceTopic` directly
        """

        if cls.reproduce_name:
            _name = cls.reproduce_name
        elif cls.reproduce_model:
            _name = cls.reproduce_model._meta.db_table
        else:
            raise ValueError("Either reproduce_name or reproduce_model must be set")

        if cls.reproduce_namespace:
            _name = f"{cls.reproduce_namespace}.{_name}"

        class _ReproduceTopic(ReproduceTopic):
            name = _name
            reproducer = cls()

        return _ReproduceTopic()

    def reproduce(self, key: Any, value: Any, is_deletion: bool) -> None:
        """receives the consumed message data and reproduces it as needed

        calls `_reproduce_upsert` or `_reproduce_deletion`, which must be implemented
        """

        if is_deletion:
            self._reproduce_deletion(key["id"], key, value)
            return

        # suppress; upsert messages aren't relevant to reproduce if the instance is gone
        with contextlib.suppress(self.reproduce_model.DoesNotExist):
            instance = self.reproduce_model.objects.get(id=key["id"])
            self._reproduce_upsert(instance, key, value)

    def _reproduce_upsert(self, instance, key, value):
        """produce message for instance upsert"""
        raise NotImplementedError

    def _reproduce_deletion(self, instance_id: int, key, value):
        """produce message for instance deletion"""
        raise NotImplementedError



import contextlib
from typing import TYPE_CHECKING, Generic, TypeVar

from django_kafka.exceptions import DjangoKafkaError, TopicNotRegisteredError

if TYPE_CHECKING:
    from django_kafka.connect.connector import Connector  # noqa: F401
    from django_kafka.consumer import Consumer  # noqa: F401


T = TypeVar("T")


class Registry(Generic[T]):
    def __init__(self):
        self._classes: dict[str, type[T]] = {}

    def __call__(self):
        def add_to_registry(cls: type[T]) -> type[T]:
            self.register(cls)
            return cls

        return add_to_registry

    def __getitem__(self, key: str):
        try:
            return self._classes[key]
        except KeyError as error:
            raise DjangoKafkaError(f"`{key}` is not registered.") from error

    def __iter__(self):
        yield from self._classes.keys()

    def __contains__(self, item: T) -> bool:
        return item in self._classes.values()

    def get_key(self, cls: type[T]) -> str:
        return f"{cls.__module__}.{cls.__name__}"

    def register(self, cls: type[T]):
        key = self.get_key(cls)
        if key in self._classes:
            raise DjangoKafkaError(f"`{key}` is already registered.")
        self._classes[key] = cls


class ConnectorsRegistry(Registry["Connector"]):
    def get_key(self, cls) -> str:
        return cls.name


class ConsumersRegistry(Registry["Consumer"]):
    def register(self, cls):
        from django_kafka.retry.consumer import RetryConsumer

        super().register(cls)

        if retry_consumer_cls := RetryConsumer.build(cls):
            self._classes[f"{self.get_key(cls)}.retry"] = retry_consumer_cls

    def topic(self, topic_name: str):
        for key in self:
            with contextlib.suppress(KeyError, TopicNotRegisteredError):
                return self[key].topics.get(topic_name=topic_name)
        return None



from django_kafka.utils.message import Header


class DeadLetterHeader(Header):
    MESSAGE_TOPIC = "DEAD_LETTER_MESSAGE_TOPIC"
    MESSAGE_PARTITION = "DEAD_LETTER_MESSAGE_PARTITION"
    MESSAGE_OFFSET = "DEAD_LETTER_MESSAGE_OFFSET"
    SUMMARY = "DEAD_LETTER_SUMMARY"
    DETAIL = "DEAD_LETTER_DETAIL"



from abc import ABC, abstractmethod
from pydoc import locate
from typing import TYPE_CHECKING

from django_kafka import kafka
from django_kafka.conf import settings
from django_kafka.exceptions import DjangoKafkaError

if TYPE_CHECKING:
    from django_kafka.relations_resolver.relation import Relation


class RelationResolverDaemon(ABC):
    @abstractmethod
    def aresolve_relation(self, relation: "Relation"): ...

    async def aresolve_relations(self):
        async for relation in kafka.relations_resolver.processor.ato_resolve():
            await self._aresolve_relation(relation)

    async def _aresolve_relation(self, relation: "Relation"):
        await relation.amark_resolving()
        await self.aresolve_relation(relation)


def get_daemon():
    if not (daemon_cls := locate(settings.RELATION_RESOLVER_DAEMON)):
        raise DjangoKafkaError(f"{settings.RELATION_RESOLVER_DAEMON} not found.")
    return daemon_cls()



# Generated by Django 5.2 on 2025-07-03 12:27

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_kafka', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='WaitingMessage',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('status', models.IntegerField(choices=[(1, 'Waiting'), (2, 'Resolving'), (3, 'Resolved')], default=1, verbose_name='status')),
                ('key', models.BinaryField(verbose_name='key')),
                ('value', models.BinaryField(verbose_name='value')),
                ('timestamp', models.JSONField(verbose_name='kafka timestamp')),
                ('topic', models.TextField(verbose_name='topic')),
                ('partition', models.TextField(verbose_name='partition')),
                ('offset', models.TextField(verbose_name='offset')),
                ('_headers', models.JSONField(null=True, verbose_name='headers')),
                ('relation_model_key', models.CharField(max_length=255, verbose_name='relation model key')),
                ('relation_id_field', models.CharField(max_length=256, verbose_name='relation id field')),
                ('relation_id_value', models.CharField(max_length=256, verbose_name='relation id value')),
                ('serialized_relation', models.JSONField(verbose_name='relation kwargs')),
            ],
            options={
                'verbose_name': 'waiting message',
                'verbose_name_plural': 'waiting messages',
                'ordering': ('offset',),
            },
        ),
    ]



from django.test import TestCase

from django_kafka.relations_resolver.relation import ModelRelation
from example.models import Order


# ruff: noqa: DJ008
class ModelRelationTestCase(TestCase):
    def test__init__(self):
        relation = ModelRelation(Order, id_field="id", id_value=1)

        self.assertEqual(relation.model, Order)
        self.assertEqual(relation.id_field, "id")
        self.assertEqual(relation.id_value, 1)
        self.assertEqual(
            relation.model_key,
            ModelRelation.get_model_key(Order),
        )

    def test_identifier(self):
        relation = ModelRelation(Order, id_field="id", id_value=1)

        self.assertEqual(
            relation.identifier(),
            f"{relation.model_key}-{relation.id_value}".lower(),
        )

    async def test_aexists(self):
        relation = ModelRelation(Order, id_field="id", id_value=1)
        self.assertFalse(await relation.aexists())

        await Order.objects.acreate(id=1)
        self.assertTrue(await relation.aexists())

    def test_serialize(self):
        relation = ModelRelation(Order, id_field="id", id_value=1)
        self.assertDictEqual(
            relation.serialize(),
            {
                "type": "MODEL",
                "kwargs": {
                    "model": ModelRelation.get_model_key(Order),
                    "id_value": 1,
                    "id_field": "id",
                },
            },
        )

    def test_deserialize(self):
        serialized_relation = {
            "type": "MODEL",
            "kwargs": {
                "model": ModelRelation.get_model_key(Order),
                "id_value": 1,
                "id_field": "id",
            },
        }
        relation = ModelRelation.deserialize(**serialized_relation["kwargs"])

        self.assertEqual(relation.model, Order)
        self.assertEqual(relation.id_field, serialized_relation["kwargs"]["id_field"])
        self.assertEqual(relation.id_value, serialized_relation["kwargs"]["id_value"])
        self.assertEqual(relation.model_key, serialized_relation["kwargs"]["model"])



from collections.abc import Iterator
from datetime import datetime
from typing import TYPE_CHECKING

from confluent_kafka import TopicPartition
from django.utils import timezone

if TYPE_CHECKING:
    from confluent_kafka import cimpl


class PauseManager:
    """Manager for partition pauses"""

    __pauses: dict[TopicPartition, datetime]

    def __init__(self):
        self.__pauses = {}

    @staticmethod
    def get_msg_partition(msg: "cimpl.Message") -> TopicPartition:
        """returns the message topic partition"""
        return TopicPartition(msg.topic(), msg.partition(), msg.offset())

    def set(self, msg: "cimpl.Message", until: datetime) -> TopicPartition:
        """adds message partition to the pause list, returning the partition"""
        tp = self.get_msg_partition(msg)
        self.__pauses[tp] = until
        return tp

    def pop_ready(self) -> Iterator[TopicPartition]:
        """returns the partitions ready to resume, removing them from the pause list"""
        now = timezone.now()
        for tp, pause in list(self.__pauses.items()):
            if now >= pause:
                del self.__pauses[tp]
                yield tp

    def reset(self):
        self.__pauses = {}


class RetryManager:
    """Manager for blocking message retry attempts"""

    __retries: dict[TopicPartition, int]

    def __init__(self):
        self.__retries = {}

    @staticmethod
    def get_msg_partition(msg: "cimpl.Message") -> TopicPartition:
        """returns the message topic partition, set to the message offset

        Note: TopicPartition hashes based on topic/partition, but not the offset.
        """
        return TopicPartition(msg.topic(), msg.partition(), msg.offset())

    def next(self, msg: "cimpl.Message"):
        """increments and returns the partition attempt count"""
        msg_tp = self.get_msg_partition(msg)
        for tp in self.__retries:
            if tp == msg_tp and tp.offset != msg_tp.offset:
                del self.__retries[tp]  # new offset encountered, reset entry
                break

        next_attempt = self.__retries.get(msg_tp, 0) + 1
        self.__retries[msg_tp] = next_attempt
        return next_attempt

    def reset(self):
        self.__retries = {}



# Generated by Django 5.1.6 on 2025-03-31 14:06

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='KeyOffsetTracker',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('topic', models.CharField(max_length=256, verbose_name='topic')),
                ('key', models.BinaryField(verbose_name='key')),
                ('offset', models.PositiveIntegerField(verbose_name='offset')),
                ('create_time', models.DateTimeField(null=True, verbose_name='create time')),
                ('log_append_time', models.DateTimeField(null=True, verbose_name='log append time')),
            ],
            options={
                'verbose_name': 'key offset tracker',
                'verbose_name_plural': 'key offsets tracker',
                'indexes': [models.Index(fields=['topic', 'key'], include=('offset',), name='django_kafka_key_offset_index')],
            },
        ),
    ]



from collections.abc import AsyncIterator
from typing import TYPE_CHECKING

from asgiref.sync import sync_to_async

from django_kafka import kafka
from django_kafka.exceptions import TopicNotRegisteredError
from django_kafka.relations_resolver.processor.base import MessageProcessor
from django_kafka.relations_resolver.relation import Relation
from django_kafka.utils.message import Message

if TYPE_CHECKING:
    from confluent_kafka import cimpl

    from django_kafka.topic import TopicConsumer


class ModelMessageProcessor(MessageProcessor):
    def __init__(self):
        from django_kafka.models import WaitingMessage

        self.model = WaitingMessage

    async def aadd_message(self, msg: "cimpl.Message", relation: "Relation"):
        await sync_to_async(self.model.objects.add_message)(msg, relation)

    async def adelete(self, relation: "Relation"):
        async_qs = await sync_to_async(self.model.objects.for_relation)(relation)
        await async_qs.adelete()

    async def aexists(self, relation: "Relation") -> bool:
        return await (
            await sync_to_async(self.model.objects.for_relation)(relation)
        ).aexists()

    async def aprocess_messages(self, relation: "Relation"):
        async for msg in await sync_to_async(self.model.objects.for_relation)(relation):
            kafka_msg = Message(
                key=msg.key,
                value=msg.value,
                topic=msg.topic,
                headers=msg.headers,
                timestamp=msg.timestamp,
                partition=msg.partition,
                offset=msg.offset,
            )
            try:
                topic: TopicConsumer = kafka.consumers.topic(kafka_msg.topic())
            except TopicNotRegisteredError:
                await msg.adelete()
                return

            if missing_relation := await self._aget_missing_relation(topic, kafka_msg):
                # doublecheck if there are other relations still missing
                # and send message for waiting into another queue
                await kafka.relations_resolver.await_for_relation(
                    kafka_msg,
                    missing_relation,
                )
                await msg.adelete()
                return

            await sync_to_async(
                kafka.consumers.topic(kafka_msg.topic()).consume,
            )(kafka_msg)
            await msg.adelete()

    async def _aget_missing_relation(
        self,
        topic: "TopicConsumer",
        msg: "cimpl.Message",
    ) -> Relation | None:
        for relation in topic.get_relations(msg):
            if not await relation.aexists():
                return relation
        return None

    async def ato_resolve(self) -> AsyncIterator["Relation"]:
        async for item in self.model.objects.aiter_relations_to_resolve(chunk_size=500):
            relation = await sync_to_async(item.relation)()
            if await relation.aexists():
                yield relation

    async def amark_resolving(self, relation: "Relation"):
        await sync_to_async(self.model.objects.mark_resolving)(relation)



from datetime import UTC, datetime
from typing import ClassVar
from unittest.mock import MagicMock, Mock, call, patch

from django.test import SimpleTestCase, TestCase
from faker import Faker

from django_kafka import kafka
from django_kafka.consumer import Consumer, Topics
from django_kafka.exceptions import DjangoKafkaError
from django_kafka.models import KeyOffsetTracker
from django_kafka.retry.tracker import KeyOffsetTrackerConsumer, KeyOffsetTrackerTopic
from django_kafka.tests.utils import message_mock
from django_kafka.topic import TopicConsumer
from django_kafka.utils.message import MessageTimestamp


class KeyOffsetTrackerConsumerTestCase(SimpleTestCase):
    @patch.multiple(
        "django_kafka.retry.tracker.KeyOffsetTrackerConsumer",
        config={},
        topics=None,
    )
    @patch("django_kafka.retry.tracker.kafka.consumers")
    @patch("django_kafka.retry.tracker.KeyOffsetTrackerConsumer.run_sanity_checks")
    @patch(
        "django_kafka.retry.tracker.KeyOffsetTrackerConsumer.tracked_topics",
        return_value=[Mock()],
    )
    def test_enable(
        self,
        mock_tracked_topics,
        mock_run_sanity_checks,
        mock_kafka_consumers,
    ):
        manager = MagicMock()
        manager.attach_mock(mock_tracked_topics, "tracked_topics")
        manager.attach_mock(mock_run_sanity_checks, "run_sanity_checks")
        manager.attach_mock(mock_kafka_consumers.return_value, "kafka_consumers")

        provided_group_id = "offset-tracker"
        self.assertIsNone(KeyOffsetTrackerConsumer.config.get("group.id"))
        KeyOffsetTrackerConsumer.activate(provided_group_id)
        mock_tracked_topics.assert_called_with()
        self.assertEqual(
            KeyOffsetTrackerConsumer.config.get("group.id"),
            provided_group_id,
        )
        self.assertEqual(
            KeyOffsetTrackerConsumer.build_config().get("group.id"),
            provided_group_id,
        )
        for topic in mock_tracked_topics.return_value:
            self.assertIn(topic, KeyOffsetTrackerConsumer.topics)

        mock_tracked_topics.assert_called_once_with()
        mock_run_sanity_checks.assert_called_once_with()
        mock_kafka_consumers.return_value.assert_called_once_with(
            KeyOffsetTrackerConsumer,
        )

        manager.assert_has_calls(
            [
                call.tracked_topics(),
                call.run_sanity_checks(),
                call.kafka_consumers(KeyOffsetTrackerConsumer),
            ],
            any_order=False,
        )

    def test_sanity_checks(self):
        no_group_id_msg = "'group.id' must be specified."
        no_topics_msg = (
            "Key offset tracker consumer is registered but there are no "
            "topics configured to use it."
        )
        with patch.multiple(
            "django_kafka.retry.tracker.KeyOffsetTrackerConsumer",
            config={"group.id": "project-key-offset-tracker"},
            topics=Topics(Mock()),
        ):
            # all required params are set, no errors raised
            KeyOffsetTrackerConsumer.run_sanity_checks()

        # nothing is set
        with self.assertRaises(DjangoKafkaError):
            KeyOffsetTrackerConsumer.run_sanity_checks()

        # group.id is not set
        with (
            self.assertRaisesMessage(DjangoKafkaError, no_group_id_msg),
            patch.multiple(
                "django_kafka.retry.tracker.KeyOffsetTrackerConsumer",
                topics=Topics(Mock()),
            ),
        ):
            KeyOffsetTrackerConsumer.run_sanity_checks()

            # topics are not set
            with (
                self.assertRaisesMessage(DjangoKafkaError, no_topics_msg),
                patch.multiple(
                    "django_kafka.retry.tracker.KeyOffsetTrackerConsumer",
                    config={"group.id": "project-key-offset-tracker"},
                ),
            ):
                KeyOffsetTrackerConsumer.run_sanity_checks()

        # topics are empty
        with (
            self.assertRaisesMessage(DjangoKafkaError, no_topics_msg),
            patch.multiple(
                "django_kafka.retry.tracker.KeyOffsetTrackerConsumer",
                config={"group.id": "project-key-offset-tracker"},
                topics=Topics(),
            ),
        ):
            KeyOffsetTrackerConsumer.run_sanity_checks()

    @patch.multiple(
        "django_kafka.retry.tracker.KeyOffsetTrackerConsumer",
        config={"group.id": "project-key-offset-tracker"},
    )
    @patch("django_kafka.consumer.consumer.ConfluentConsumer", new=Mock())
    @patch("django_kafka.retry.tracker.KeyOffsetTrackerConsumer.run_sanity_checks")
    @patch("django_kafka.retry.tracker.super")
    def start_runs_sanity_checks(self, mock_super, mock_run_sanity_checks):
        manager = MagicMock()
        manager.attach_mock(mock_super().start, "super")
        manager.attach_mock(mock_run_sanity_checks, "run_sanity_checks")

        KeyOffsetTrackerConsumer().start()

        mock_run_sanity_checks.assert_called_once_with()
        mock_super().start.assert_called_once_with()

        manager.assert_has_calls(
            [
                call.run_sanity_checks(),
                call.super(),
            ],
            any_order=False,
        )

    def test_tracked_topics(self):
        @kafka.retry(max_retries=3, delay=120, use_offset_tracker=True)
        class Topic1(TopicConsumer):
            name = "topic-1"

            def consume(self, msg):
                pass

        @kafka.retry(max_retries=3, delay=120, use_offset_tracker=True)
        class Topic2(TopicConsumer):
            name = "topic-2"

            def consume(self, msg):
                pass

        class SimpleTopic(TopicConsumer):
            name = "topic-3"

            def consume(self, msg):
                pass

        @kafka.consumers()
        class Consumer1(Consumer):
            config: ClassVar = {"group.id": "group_id1"}
            topics = Topics(Topic1(), SimpleTopic())

        @kafka.consumers()
        class Consumer2(Consumer):
            config: ClassVar = {"group.id": "group_id2"}
            topics = Topics(Topic2(), Topic1())

        topics_using_key_offset = (Topic1, Topic2)
        topics_not_using_key_offset = [SimpleTopic]

        tracked_topics = KeyOffsetTrackerConsumer.tracked_topics()

        self.assertListEqual(
            sorted([topic.name for topic in topics_using_key_offset]),
            sorted([topic.name for topic in tracked_topics]),
        )

        for topic in topics_not_using_key_offset:
            self.assertNotIn(topic.name, [t.name for t in tracked_topics])

        for topic in KeyOffsetTrackerConsumer.tracked_topics():
            self.assertIsInstance(topic, KeyOffsetTrackerTopic)


class TestKeyOffsetTrackerTopic(SimpleTestCase):
    def test_init_sets_topic_name(self):
        name = "topic-name"
        topic = KeyOffsetTrackerTopic(name)
        self.assertEqual(topic.name, name)

    @patch("django_kafka.models.KeyOffsetTracker.objects.log_msg_offset")
    def test_consume_logs_to_db(self, mock_log_msg_offset):
        msg = Mock()
        KeyOffsetTrackerTopic("topic-name").consume(msg)
        mock_log_msg_offset.assert_called_once_with(msg)


class KeyOffsetTrackerQuerySetTestCase(TestCase):
    @classmethod
    def setUpClass(cls):
        cls.faker = Faker()
        super().setUpClass()

    def test_log_msg_offset(self):
        now = datetime.now(UTC)
        timestamp = now.timestamp() * 1000
        msg1 = message_mock(
            offset=1,
            timestamp=[MessageTimestamp.CREATE_TIME, timestamp],
        )

        qs = KeyOffsetTracker.objects.all()

        instance, created = KeyOffsetTracker.objects.log_msg_offset(msg1)
        self.assertTrue(created)
        self.assertEqual(instance.offset, 1)
        self.assertEqual(instance.create_time, now)
        self.assertIsNone(instance.log_append_time)
        self.assertEqual(qs.count(), 1)

        msg1.offset.return_value = 2
        instance, created = KeyOffsetTracker.objects.log_msg_offset(msg1)
        self.assertFalse(created)
        self.assertEqual(instance.offset, 2)
        self.assertEqual(instance.create_time, now)
        self.assertIsNone(instance.log_append_time)
        self.assertEqual(qs.count(), 1)

        msg2 = message_mock(offset=100)
        instance, created = KeyOffsetTracker.objects.log_msg_offset(msg2)
        self.assertTrue(created)
        self.assertEqual(instance.offset, 100)
        self.assertIsNone(instance.create_time)
        self.assertIsNone(instance.log_append_time)
        self.assertEqual(qs.count(), 2)

    def test_has_future_offset(self):
        msg = message_mock()

        KeyOffsetTracker.objects.create(
            topic=msg.topic(),
            key=msg.key(),
            offset=msg.offset(),
        )
        self.assertFalse(KeyOffsetTracker.objects.has_future_offset(msg))

        KeyOffsetTracker.objects.update(offset=msg.offset() + 100)
        self.assertTrue(KeyOffsetTracker.objects.has_future_offset(msg))


class KeyOffsetTrackerModelTestCase(TestCase):
    def test_timestamp_property(self):
        now = datetime.now(UTC)
        timestamp = now.timestamp() * 1000
        instance = KeyOffsetTracker()

        instance.timestamp = [MessageTimestamp.CREATE_TIME, timestamp]
        self.assertEqual(instance.create_time, now)
        self.assertIsNone(instance.log_append_time)

        instance.timestamp = [MessageTimestamp.LOG_APPEND_TIME, timestamp]
        self.assertEqual(instance.log_append_time, now)
        self.assertIsNone(instance.create_time)

        instance.timestamp = [MessageTimestamp.NOT_AVAILABLE, timestamp]
        self.assertIsNone(instance.create_time)
        self.assertIsNone(instance.log_append_time)
